"""
Sample models and functions prepared for use in mystic


Functions
=========

    corana 1d,2d,3d,4d     -- Corana's function


Models
======

    dense_circle     -- 2d array representation of a circle
    sparse_circle
    minimal_circle
    decay            -- Bevington & Robinson's model of dual exponential decay

"""

# base classes
# from .model import Fitness, Function

# models
# from .decay import decay
# from .circle import dense_circle, sparse_circle, minimal_circle

# functions
# from .corana import corana1d, corana2d, corana3d, corana4d

# end of file
